@@ -33,163 +33,44 @@
 This file defines container classes which represent categories of protocol
 buffer field types which need extra maintenance. Currently these categories
 are:
-  - Repeated scalar fields - These are all repeated fields which aren't
+
+-   Repeated scalar fields - These are all repeated fields which aren't
     composite (e.g. they are of simple types like int32, string, etc).
-  - Repeated composite fields - Repeated fields which are composite. This
+-   Repeated composite fields - Repeated fields which are composite. This
     includes groups and nested messages.
 """
 
-__author__ = 'petar@google.com (Petar Petrov)'
-
-import collections
-import sys
-
-if sys.version_info[0] < 3:
-  # We would use collections.MutableMapping all the time, but in Python 2 it
-  # doesn't define __slots__.  This causes two significant problems:
-  #
-  # 1. we can't disallow arbitrary attribute assignment, even if our derived
-  #    classes *do* define __slots__.
-  #
-  # 2. we can't safely derive a C type from it without __slots__ defined (the
-  #    interpreter expects to find a dict at tp_dictoffset, which we can't
-  #    robustly provide.  And we don't want an instance dict anyway.
-  #
-  # So this is the Python 2.7 definition of Mapping/MutableMapping functions
-  # verbatim, except that:
-  # 1. We declare __slots__.
-  # 2. We don't declare this as a virtual base class.  The classes defined
-  #    in collections are the interesting base classes, not us.
-  #
-  # Note: deriving from object is critical.  It is the only thing that makes
-  # this a true type, allowing us to derive from it in C++ cleanly and making
-  # __slots__ properly disallow arbitrary element assignment.
-
-  class Mapping(object):
-    __slots__ = ()
-
-    def get(self, key, default=None):
-      try:
-        return self[key]
-      except KeyError:
-        return default
-
-    def __contains__(self, key):
-      try:
-        self[key]
-      except KeyError:
-        return False
-      else:
-        return True
-
-    def iterkeys(self):
-      return iter(self)
-
-    def itervalues(self):
-      for key in self:
-        yield self[key]
-
-    def iteritems(self):
-      for key in self:
-        yield (key, self[key])
-
-    def keys(self):
-      return list(self)
-
-    def items(self):
-      return [(key, self[key]) for key in self]
-
-    def values(self):
-      return [self[key] for key in self]
-
-    # Mappings are not hashable by default, but subclasses can change this
-    __hash__ = None
-
-    def __eq__(self, other):
-      if not isinstance(other, collections.Mapping):
-        return NotImplemented
-      return dict(list(self.items())) == dict(list(other.items()))
-
-    def __ne__(self, other):
-      return not (self == other)
-
-  class MutableMapping(Mapping):
-    __slots__ = ()
-
-    __marker = object()
-
-    def pop(self, key, default=__marker):
-      try:
-        value = self[key]
-      except KeyError:
-        if default is self.__marker:
-          raise
-        return default
-      else:
-        del self[key]
-        return value
-
-    def popitem(self):
-      try:
-        key = next(iter(self))
-      except StopIteration:
-        raise KeyError
-      value = self[key]
-      del self[key]
-      return key, value
-
-    def clear(self):
-      try:
-        while True:
-          self.popitem()
-      except KeyError:
-        pass
-
-    def update(*args, **kwds):
-      if len(args) > 2:
-        raise TypeError("update() takes at most 2 positional "
-                        "arguments ({} given)".format(len(args)))
-      elif not args:
-        raise TypeError("update() takes at least 1 argument (0 given)")
-      self = args[0]
-      other = args[1] if len(args) >= 2 else ()
-
-      if isinstance(other, Mapping):
-        for key in other:
-          self[key] = other[key]
-      elif hasattr(other, "keys"):
-        for key in list(other.keys()):
-          self[key] = other[key]
-      else:
-        for key, value in other:
-          self[key] = value
-      for key, value in list(kwds.items()):
-        self[key] = value
-
-    def setdefault(self, key, default=None):
-      try:
-        return self[key]
-      except KeyError:
-        self[key] = default
-      return default
-
-  collections.Mapping.register(Mapping)
-  collections.MutableMapping.register(MutableMapping)
-
-else:
-  # In Python 3 we can just use MutableMapping directly, because it defines
-  # __slots__.
-  MutableMapping = collections.MutableMapping
+import collections.abc
+import copy
+import pickle
+from typing import (
+    Any,
+    Iterable,
+    Iterator,
+    List,
+    MutableMapping,
+    MutableSequence,
+    NoReturn,
+    Optional,
+    Sequence,
+    TypeVar,
+    Union,
+    overload,
+)
 
 
-class BaseContainer(object):
+_T = TypeVar('_T')
+_K = TypeVar('_K')
+_V = TypeVar('_V')
 
+
+class BaseContainer(Sequence[_T]):
   """Base container class."""
 
   # Minimizes memory usage and disallows assignment to other attributes.
   __slots__ = ['_message_listener', '_values']
 
-  def __init__(self, message_listener):
+  def __init__(self, message_listener: Any) -> None:
     """
     Args:
       message_listener: A MessageListener implementation.
@@ -199,26 +80,33 @@ class BaseContainer(object):
     self._message_listener = message_listener
     self._values = []
 
+  @overload
+  def __getitem__(self, key: int) -> _T:
+    ...
+
+  @overload
+  def __getitem__(self, key: slice) -> List[_T]:
+    ...
+
   def __getitem__(self, key):
     """Retrieves item by the specified key."""
     return self._values[key]
 
-  def __len__(self):
+  def __len__(self) -> int:
     """Returns the number of elements in the container."""
     return len(self._values)
 
-  def __ne__(self, other):
+  def __ne__(self, other: Any) -> bool:
     """Checks if another instance isn't equal to this one."""
     # The concrete classes should define __eq__.
     return not self == other
 
-  def __hash__(self):
-    raise TypeError('unhashable object')
+  __hash__ = None
 
-  def __repr__(self):
+  def __repr__(self) -> str:
     return repr(self._values)
 
-  def sort(self, *args, **kwargs):
+  def sort(self, *args, **kwargs) -> None:
     # Continue to support the old sort_function keyword argument.
     # This is expected to be a rare occurrence, so use LBYL to avoid
     # the overhead of actually catching KeyError.
@@ -226,41 +114,51 @@ class BaseContainer(object):
       kwargs['cmp'] = kwargs.pop('sort_function')
     self._values.sort(*args, **kwargs)
 
+  def reverse(self) -> None:
+    self._values.reverse()
 
-class RepeatedScalarFieldContainer(BaseContainer):
 
+# TODO(slebedev): Remove this. BaseContainer does *not* conform to
+# MutableSequence, only its subclasses do.
+collections.abc.MutableSequence.register(BaseContainer)
+
+
+class RepeatedScalarFieldContainer(BaseContainer[_T], MutableSequence[_T]):
   """Simple, type-checked, list-like container for holding repeated scalars."""
 
   # Disallows assignment to other attributes.
   __slots__ = ['_type_checker']
 
-  def __init__(self, message_listener, type_checker):
-    """
-    Args:
-      message_listener: A MessageListener implementation.
-        The RepeatedScalarFieldContainer will call this object's
-        Modified() method when it is modified.
+  def __init__(
+      self,
+      message_listener: Any,
+      type_checker: Any,
+  ) -> None:
+    """Args:
+
+      message_listener: A MessageListener implementation. The
+      RepeatedScalarFieldContainer will call this object's Modified() method
+      when it is modified.
       type_checker: A type_checkers.ValueChecker instance to run on elements
-        inserted into this container.
+      inserted into this container.
     """
-    super(RepeatedScalarFieldContainer, self).__init__(message_listener)
+    super().__init__(message_listener)
     self._type_checker = type_checker
 
-  def append(self, value):
+  def append(self, value: _T) -> None:
     """Appends an item to the list. Similar to list.append()."""
     self._values.append(self._type_checker.CheckValue(value))
     if not self._message_listener.dirty:
       self._message_listener.Modified()
 
-  def insert(self, key, value):
+  def insert(self, key: int, value: _T) -> None:
     """Inserts the item at the specified position. Similar to list.insert()."""
     self._values.insert(key, self._type_checker.CheckValue(value))
     if not self._message_listener.dirty:
       self._message_listener.Modified()
 
-  def extend(self, elem_seq):
+  def extend(self, elem_seq: Iterable[_T]) -> None:
     """Extends by appending the given iterable. Similar to list.extend()."""
-
     if elem_seq is None:
       return
     try:
@@ -277,57 +175,52 @@ class RepeatedScalarFieldContainer(BaseContainer):
       self._values.extend(new_values)
     self._message_listener.Modified()
 
-  def MergeFrom(self, other):
+  def MergeFrom(
+      self,
+      other: Union['RepeatedScalarFieldContainer[_T]', Iterable[_T]],
+  ) -> None:
     """Appends the contents of another repeated field of the same type to this
     one. We do not check the types of the individual fields.
     """
-    self._values.extend(other._values)
+    self._values.extend(other)
     self._message_listener.Modified()
 
-  def remove(self, elem):
+  def remove(self, elem: _T):
     """Removes an item from the list. Similar to list.remove()."""
     self._values.remove(elem)
     self._message_listener.Modified()
 
-  def pop(self, key=-1):
+  def pop(self, key: Optional[int] = -1) -> _T:
     """Removes and returns an item at a given index. Similar to list.pop()."""
     value = self._values[key]
     self.__delitem__(key)
     return value
 
-  def __setitem__(self, key, value):
+  @overload
+  def __setitem__(self, key: int, value: _T) -> None:
+    ...
+
+  @overload
+  def __setitem__(self, key: slice, value: Iterable[_T]) -> None:
+    ...
+
+  def __setitem__(self, key, value) -> None:
     """Sets the item on the specified position."""
-    if isinstance(key, slice):  # PY3
+    if isinstance(key, slice):
       if key.step is not None:
         raise ValueError('Extended slices not supported')
-      self.__setslice__(key.start, key.stop, value)
+      self._values[key] = map(self._type_checker.CheckValue, value)
+      self._message_listener.Modified()
     else:
       self._values[key] = self._type_checker.CheckValue(value)
       self._message_listener.Modified()
 
-  def __getslice__(self, start, stop):
-    """Retrieves the subset of items from between the specified indices."""
-    return self._values[start:stop]
-
-  def __setslice__(self, start, stop, values):
-    """Sets the subset of items from between the specified indices."""
-    new_values = []
-    for value in values:
-      new_values.append(self._type_checker.CheckValue(value))
-    self._values[start:stop] = new_values
-    self._message_listener.Modified()
-
-  def __delitem__(self, key):
+  def __delitem__(self, key: Union[int, slice]) -> None:
     """Deletes the item at the specified position."""
     del self._values[key]
     self._message_listener.Modified()
 
-  def __delslice__(self, start, stop):
-    """Deletes the subset of items from between the specified indices."""
-    del self._values[start:stop]
-    self._message_listener.Modified()
-
-  def __eq__(self, other):
+  def __eq__(self, other: Any) -> bool:
     """Compares the current instance with another one."""
     if self is other:
       return True
@@ -337,17 +230,28 @@ class RepeatedScalarFieldContainer(BaseContainer):
     # We are presumably comparing against some other sequence type.
     return other == self._values
 
-collections.MutableSequence.register(BaseContainer)
+  def __deepcopy__(
+      self,
+      unused_memo: Any = None,
+  ) -> 'RepeatedScalarFieldContainer[_T]':
+    clone = RepeatedScalarFieldContainer(
+        copy.deepcopy(self._message_listener), self._type_checker)
+    clone.MergeFrom(self)
+    return clone
+
+  def __reduce__(self, **kwargs) -> NoReturn:
+    raise pickle.PickleError(
+        "Can't pickle repeated scalar fields, convert to list first")
 
 
-class RepeatedCompositeFieldContainer(BaseContainer):
-
+# TODO(slebedev): Constrain T to be a subtype of Message.
+class RepeatedCompositeFieldContainer(BaseContainer[_T], MutableSequence[_T]):
   """Simple, list-like container for holding repeated composite fields."""
 
   # Disallows assignment to other attributes.
   __slots__ = ['_message_descriptor']
 
-  def __init__(self, message_listener, message_descriptor):
+  def __init__(self, message_listener: Any, message_descriptor: Any) -> None:
     """
     Note that we pass in a descriptor instead of the generated directly,
     since at the time we construct a _RepeatedCompositeFieldContainer we
@@ -362,10 +266,10 @@ class RepeatedCompositeFieldContainer(BaseContainer):
         that should be present in this container.  We'll use the
         _concrete_class field of this descriptor when the client calls add().
     """
-    super(RepeatedCompositeFieldContainer, self).__init__(message_listener)
+    super().__init__(message_listener)
     self._message_descriptor = message_descriptor
 
-  def add(self, **kwargs):
+  def add(self, **kwargs: Any) -> _T:
     """Adds a new element at the end of the list and returns it. Keyword
     arguments may be used to initialize the element.
     """
@@ -376,8 +280,27 @@ class RepeatedCompositeFieldContainer(BaseContainer):
       self._message_listener.Modified()
     return new_element
 
-  def extend(self, elem_seq):
+  def append(self, value: _T) -> None:
+    """Appends one element by copying the message."""
+    new_element = self._message_descriptor._concrete_class()
+    new_element._SetListener(self._message_listener)
+    new_element.CopyFrom(value)
+    self._values.append(new_element)
+    if not self._message_listener.dirty:
+      self._message_listener.Modified()
+
+  def insert(self, key: int, value: _T) -> None:
+    """Inserts the item at the specified position by copying."""
+    new_element = self._message_descriptor._concrete_class()
+    new_element._SetListener(self._message_listener)
+    new_element.CopyFrom(value)
+    self._values.insert(key, new_element)
+    if not self._message_listener.dirty:
+      self._message_listener.Modified()
+
+  def extend(self, elem_seq: Iterable[_T]) -> None:
     """Extends by appending the given sequence of elements of the same type
+
     as this one, copying each individual message.
     """
     message_class = self._message_descriptor._concrete_class
@@ -390,38 +313,47 @@ class RepeatedCompositeFieldContainer(BaseContainer):
       values.append(new_element)
     listener.Modified()
 
-  def MergeFrom(self, other):
+  def MergeFrom(
+      self,
+      other: Union['RepeatedCompositeFieldContainer[_T]', Iterable[_T]],
+  ) -> None:
     """Appends the contents of another repeated field of the same type to this
     one, copying each individual message.
     """
-    self.extend(other._values)
+    self.extend(other)
 
-  def remove(self, elem):
+  def remove(self, elem: _T) -> None:
     """Removes an item from the list. Similar to list.remove()."""
     self._values.remove(elem)
     self._message_listener.Modified()
 
-  def pop(self, key=-1):
+  def pop(self, key: Optional[int] = -1) -> _T:
     """Removes and returns an item at a given index. Similar to list.pop()."""
     value = self._values[key]
     self.__delitem__(key)
     return value
 
-  def __getslice__(self, start, stop):
-    """Retrieves the subset of items from between the specified indices."""
-    return self._values[start:stop]
+  @overload
+  def __setitem__(self, key: int, value: _T) -> None:
+    ...
 
-  def __delitem__(self, key):
+  @overload
+  def __setitem__(self, key: slice, value: Iterable[_T]) -> None:
+    ...
+
+  def __setitem__(self, key, value):
+    # This method is implemented to make RepeatedCompositeFieldContainer
+    # structurally compatible with typing.MutableSequence. It is
+    # otherwise unsupported and will always raise an error.
+    raise TypeError(
+        f'{self.__class__.__name__} object does not support item assignment')
+
+  def __delitem__(self, key: Union[int, slice]) -> None:
     """Deletes the item at the specified position."""
     del self._values[key]
     self._message_listener.Modified()
 
-  def __delslice__(self, start, stop):
-    """Deletes the subset of items from between the specified indices."""
-    del self._values[start:stop]
-    self._message_listener.Modified()
-
-  def __eq__(self, other):
+  def __eq__(self, other: Any) -> bool:
     """Compares the current instance with another one."""
     if self is other:
       return True
@@ -431,16 +363,20 @@ class RepeatedCompositeFieldContainer(BaseContainer):
     return self._values == other._values
 
 
-class ScalarMap(MutableMapping):
-
+class ScalarMap(MutableMapping[_K, _V]):
   """Simple, type-checked, dict-like container for holding repeated scalars."""
 
   # Disallows assignment to other attributes.
   __slots__ = ['_key_checker', '_value_checker', '_values', '_message_listener',
                '_entry_descriptor']
 
-  def __init__(self, message_listener, key_checker, value_checker,
-               entry_descriptor):
+  def __init__(
+      self,
+      message_listener: Any,
+      key_checker: Any,
+      value_checker: Any,
+      entry_descriptor: Any,
+  ) -> None:
     """
     Args:
       message_listener: A MessageListener implementation.
@@ -458,7 +394,7 @@ class ScalarMap(MutableMapping):
     self._entry_descriptor = entry_descriptor
     self._values = {}
 
-  def __getitem__(self, key):
+  def __getitem__(self, key: _K) -> _V:
     try:
       return self._values[key]
     except KeyError:
@@ -467,12 +403,20 @@ class ScalarMap(MutableMapping):
       self._values[key] = val
       return val
 
-  def __contains__(self, item):
+  def __contains__(self, item: _K) -> bool:
     # We check the key's type to match the strong-typing flavor of the API.
     # Also this makes it easier to match the behavior of the C++ implementation.
     self._key_checker.CheckValue(item)
     return item in self._values
 
+  @overload
+  def get(self, key: _K) -> Optional[_V]:
+    ...
+
+  @overload
+  def get(self, key: _K, default: _T) -> Union[_V, _T]:
+    ...
+
   # We need to override this explicitly, because our defaultdict-like behavior
   # will make the default implementation (from our base class) always insert
   # the key.
@@ -482,30 +426,30 @@ class ScalarMap(MutableMapping):
     else:
       return default
 
-  def __setitem__(self, key, value):
+  def __setitem__(self, key: _K, value: _V) -> _T:
     checked_key = self._key_checker.CheckValue(key)
     checked_value = self._value_checker.CheckValue(value)
     self._values[checked_key] = checked_value
     self._message_listener.Modified()
 
-  def __delitem__(self, key):
+  def __delitem__(self, key: _K) -> None:
     del self._values[key]
     self._message_listener.Modified()
 
-  def __len__(self):
+  def __len__(self) -> int:
     return len(self._values)
 
-  def __iter__(self):
+  def __iter__(self) -> Iterator[_K]:
     return iter(self._values)
 
-  def __repr__(self):
+  def __repr__(self) -> str:
     return repr(self._values)
 
-  def MergeFrom(self, other):
+  def MergeFrom(self, other: 'ScalarMap[_K, _V]') -> None:
     self._values.update(other._values)
     self._message_listener.Modified()
 
-  def InvalidateIterators(self):
+  def InvalidateIterators(self) -> None:
     # It appears that the only way to reliably invalidate iterators to
     # self._values is to ensure that its size changes.
     original = self._values
@@ -513,24 +457,28 @@ class ScalarMap(MutableMapping):
     original[None] = None
 
   # This is defined in the abstract base, but we can do it much more cheaply.
-  def clear(self):
+  def clear(self) -> None:
     self._values.clear()
     self._message_listener.Modified()
 
-  def GetEntryClass(self):
+  def GetEntryClass(self) -> Any:
     return self._entry_descriptor._concrete_class
 
 
-class MessageMap(MutableMapping):
-
+class MessageMap(MutableMapping[_K, _V]):
   """Simple, type-checked, dict-like container for with submessage values."""
 
   # Disallows assignment to other attributes.
   __slots__ = ['_key_checker', '_values', '_message_listener',
                '_message_descriptor', '_entry_descriptor']
 
-  def __init__(self, message_listener, message_descriptor, key_checker,
-               entry_descriptor):
+  def __init__(
+      self,
+      message_listener: Any,
+      message_descriptor: Any,
+      key_checker: Any,
+      entry_descriptor: Any,
+  ) -> None:
     """
     Args:
       message_listener: A MessageListener implementation.
@@ -548,7 +496,7 @@ class MessageMap(MutableMapping):
     self._entry_descriptor = entry_descriptor
     self._values = {}
 
-  def __getitem__(self, key):
+  def __getitem__(self, key: _K) -> _V:
     key = self._key_checker.CheckValue(key)
     try:
       return self._values[key]
@@ -557,10 +505,9 @@ class MessageMap(MutableMapping):
       new_element._SetListener(self._message_listener)
       self._values[key] = new_element
       self._message_listener.Modified()
-
       return new_element
 
-  def get_or_create(self, key):
+  def get_or_create(self, key: _K) -> _V:
     """get_or_create() is an alias for getitem (ie. map[key]).
 
     Args:
@@ -574,6 +521,14 @@ class MessageMap(MutableMapping):
     """
     return self[key]
 
+  @overload
+  def get(self, key: _K) -> Optional[_V]:
+    ...
+
+  @overload
+  def get(self, key: _K, default: _T) -> Union[_V, _T]:
+    ...
+
   # We need to override this explicitly, because our defaultdict-like behavior
   # will make the default implementation (from our base class) always insert
   # the key.
@@ -583,29 +538,30 @@ class MessageMap(MutableMapping):
     else:
       return default
 
-  def __contains__(self, item):
+  def __contains__(self, item: _K) -> bool:
     item = self._key_checker.CheckValue(item)
     return item in self._values
 
-  def __setitem__(self, key, value):
+  def __setitem__(self, key: _K, value: _V) -> NoReturn:
     raise ValueError('May not set values directly, call my_map[key].foo = 5')
 
-  def __delitem__(self, key):
+  def __delitem__(self, key: _K) -> None:
     key = self._key_checker.CheckValue(key)
     del self._values[key]
     self._message_listener.Modified()
 
-  def __len__(self):
+  def __len__(self) -> int:
     return len(self._values)
 
-  def __iter__(self):
+  def __iter__(self) -> Iterator[_K]:
     return iter(self._values)
 
-  def __repr__(self):
+  def __repr__(self) -> str:
     return repr(self._values)
 
-  def MergeFrom(self, other):
-    for key in other:
+  def MergeFrom(self, other: 'MessageMap[_K, _V]') -> None:
+    # pylint: disable=protected-access
+    for key in other._values:
       # According to documentation: "When parsing from the wire or when merging,
       # if there are duplicate map keys the last key seen is used".
       if key in self:
@@ -614,7 +570,7 @@ class MessageMap(MutableMapping):
     # self._message_listener.Modified() not required here, because
     # mutations to submessages already propagate.
 
-  def InvalidateIterators(self):
+  def InvalidateIterators(self) -> None:
     # It appears that the only way to reliably invalidate iterators to
     # self._values is to ensure that its size changes.
     original = self._values
@@ -622,9 +578,133 @@ class MessageMap(MutableMapping):
     original[None] = None
 
   # This is defined in the abstract base, but we can do it much more cheaply.
-  def clear(self):
+  def clear(self) -> None:
     self._values.clear()
     self._message_listener.Modified()
 
-  def GetEntryClass(self):
+  def GetEntryClass(self) -> Any:
     return self._entry_descriptor._concrete_class
+
+
+class _UnknownField:
+  """A parsed unknown field."""
+
+  # Disallows assignment to other attributes.
+  __slots__ = ['_field_number', '_wire_type', '_data']
+
+  def __init__(self, field_number, wire_type, data):
+    self._field_number = field_number
+    self._wire_type = wire_type
+    self._data = data
+    return
+
+  def __lt__(self, other):
+    # pylint: disable=protected-access
+    return self._field_number < other._field_number
+
+  def __eq__(self, other):
+    if self is other:
+      return True
+    # pylint: disable=protected-access
+    return (self._field_number == other._field_number and
+            self._wire_type == other._wire_type and
+            self._data == other._data)
+
+
+class UnknownFieldRef:  # pylint: disable=missing-class-docstring
+
+  def __init__(self, parent, index):
+    self._parent = parent
+    self._index = index
+
+  def _check_valid(self):
+    if not self._parent:
+      raise ValueError('UnknownField does not exist. '
+                       'The parent message might be cleared.')
+    if self._index >= len(self._parent):
+      raise ValueError('UnknownField does not exist. '
+                       'The parent message might be cleared.')
+
+  @property
+  def field_number(self):
+    self._check_valid()
+    # pylint: disable=protected-access
+    return self._parent._internal_get(self._index)._field_number
+
+  @property
+  def wire_type(self):
+    self._check_valid()
+    # pylint: disable=protected-access
+    return self._parent._internal_get(self._index)._wire_type
+
+  @property
+  def data(self):
+    self._check_valid()
+    # pylint: disable=protected-access
+    return self._parent._internal_get(self._index)._data
+
+
+class UnknownFieldSet:
+  """UnknownField container"""
+
+  # Disallows assignment to other attributes.
+  __slots__ = ['_values']
+
+  def __init__(self):
+    self._values = []
+
+  def __getitem__(self, index):
+    if self._values is None:
+      raise ValueError('UnknownFields does not exist. '
+                       'The parent message might be cleared.')
+    size = len(self._values)
+    if index < 0:
+      index += size
+    if index < 0 or index >= size:
+      raise IndexError('index %d out of range'.index)
+
+    return UnknownFieldRef(self, index)
+
+  def _internal_get(self, index):
+    return self._values[index]
+
+  def __len__(self):
+    if self._values is None:
+      raise ValueError('UnknownFields does not exist. '
+                       'The parent message might be cleared.')
+    return len(self._values)
+
+  def _add(self, field_number, wire_type, data):
+    unknown_field = _UnknownField(field_number, wire_type, data)
+    self._values.append(unknown_field)
+    return unknown_field
+
+  def __iter__(self):
+    for i in range(len(self)):
+      yield UnknownFieldRef(self, i)
+
+  def _extend(self, other):
+    if other is None:
+      return
+    # pylint: disable=protected-access
+    self._values.extend(other._values)
+
+  def __eq__(self, other):
+    if self is other:
+      return True
+    # Sort unknown fields because their order shouldn't
+    # affect equality test.
+    values = list(self._values)
+    if other is None:
+      return not values
+    values.sort()
+    # pylint: disable=protected-access
+    other_values = sorted(other._values)
+    return values == other_values
+
+  def _clear(self):
+    for value in self._values:
+      # pylint: disable=protected-access
+      if isinstance(value._data, UnknownFieldSet):
+        value._data._clear()  # pylint: disable=protected-access
+    self._values = None
